/*
 * ArduinoExampleLL:
 *
 * Arduino sketch to present a user interface on a Reach Technology
 * SLCD43 Development Kit display, receive button press and slider
 * position notifications, and then drive an analog pulse with a 
 * settable pulse profile (PWM duty cycle and duration).
 *
 * This sketch is an example of a low-level control application
 * that manages the display completely, drawing objects on the
 * screen as required and handling all button/slider notifications
 * directly.
 *
 * The main display presents several buttons to select pre-set
 * pulse profiles.
 *
 * An AltScr button selects an alternate display screen that
 * presents two sliders to manually select the pulse profile
 * parameters.
 *
 * A TRIG button on either screen starts the pulse, as will
 * activating a short-to-ground switch connected to digital
 * pin 2.
 *
 * Timer2 on the Arduio is used to drive a PWM signal to create
 * the output pulse.  The PWM output from timer2 on OC2B (pin 9
 * on Mega2560) is filtered with an RC network to convert the
 * output into an analog signal.  The duty cycle is varied to
 * adjust the pulse amplitude.  The pulse duty cycle and duration
 * are based on information received from the display.
 *
 * On both screens, the low-level pulse profile settings are
 * displayed each time the pulse is triggered (either by the
 * TRIG button or by the footswitch).
 *
 * An LED with 330 ohm current limiting resistor connected to
 * pin 12 will turn on for the selected pulse time each time a
 * pulse is triggered.
 */

// compiler-provided definitions
// __AVR_ATmega328P__ - Uno
// __AVR_ATmega1280__ - Mega
// __AVR_ATmega2560__ - Mega 2560


// button image size literals, based on bitmap image sizes
#define  BUTTON_30_WIDTH    77    // width of 30_big_button BMP
#define  BUTTON_30_HEIGHT   62    // height of 30_big_button BMP
#define  SLIDER_44_WIDTH    81    // width of 44_slider2Back BMP
#define  SLIDER_44_HEIGHT   176   // height of 44_slider2Back BMP

// button placement literals based on button dimensions
#define  BUTTON_ROW_OFFSET  8
#define  BUTTON_COL_OFFSET  4
#define  BUTTON_SPACING     2
#define  BUTTON_ROW_1       (BUTTON_ROW_OFFSET)
#define  BUTTON_ROW_2       (BUTTON_ROW_1+BUTTON_30_HEIGHT+BUTTON_SPACING)
#define  BUTTON_ROW_3       (BUTTON_ROW_2+BUTTON_30_HEIGHT+BUTTON_SPACING)
#define  BUTTON_ROW_4       (BUTTON_ROW_3+BUTTON_30_HEIGHT+BUTTON_SPACING)
#define  BUTTON_COL_1       (BUTTON_COL_OFFSET)
#define  BUTTON_COL_2       (BUTTON_COL_1+BUTTON_30_WIDTH+BUTTON_SPACING)
#define  BUTTON_COL_3       (BUTTON_COL_2+BUTTON_30_WIDTH+BUTTON_SPACING)
#define  BUTTON_COL_4       (BUTTON_COL_3+BUTTON_30_WIDTH+BUTTON_SPACING)
#define  BUTTON_COL_5       (BUTTON_COL_4+BUTTON_30_WIDTH+BUTTON_SPACING)
#define  BUTTON_COL_6       (BUTTON_COL_5+BUTTON_30_WIDTH+BUTTON_SPACING)

// index value assignments for 'special' buttons and sliders
#define  MAIN_SCREEN_BUTTON 20
#define  ALT_SCREEN_BUTTON  21
#define  TRIGGER_BUTTON     50
#define  DUTYCYCLE_SLIDER   128
#define  DURATION_SLIDER    129

// slider image size literals
#define  SLIDER_ROW                  50
#define  SLIDER_LABEL_ROW            (SLIDER_ROW-20)
#define  DUTYCYCLE_SLIDER_COL        70
#define  DUTYCYCLE_SLIDER_LABEL_COL  (DUTYCYCLE_SLIDER_COL+5)
#define  DURATION_SLIDER_COL         230
#define  DURATION_SLIDER_LABEL_COL   (DURATION_SLIDER_COL+5)

// notifications from display, see SLCDx Software Reference
#define  BUTTON_PRESS      'x'
#define  SLIDER_ACTION     'l'

// maximum number of commands sent without getting a reply
#define  MAX_CMD_OUTSTANDING  2

// duty cycle definitions assuming 'fast PWM' config
#define  DUTY_CYCLE_MAX     255
#define  DUTY_CYCLE_20PCNT  ((DUTY_CYCLE_MAX*20)/100)
#define  DUTY_CYCLE_40PCNT  ((DUTY_CYCLE_MAX*40)/100)
#define  DUTY_CYCLE_60PCNT  ((DUTY_CYCLE_MAX*60)/100)
#define  DUTY_CYCLE_99PCNT  ((DUTY_CYCLE_MAX*99)/100)
#define  MIN_DUTY_CYCLE     DUTY_CYCLE_20PCNT

// duration literals
#define  MIN_DURATION      50      // ms
#define  MAX_DURATION      200     // ms

// length of serial input buffer
#define  BUFF_SIZE        100

// pin definitions
#define  FOOTSWITCH_IN    2
#define  PWM_OUT          9  // timer2 OC2B output
#define  DEBUG_LED        12

// serial IO port mapping literals
#define  DISPLAY_SIO          Serial1
#define  DISPLAY_SERIALEVENT  serialEvent1
#define  DEBUG_SIO            Serial    // to use IDE's Serial Monitor

// debug support
#undef  SHOW_DEBUG
#ifdef  SHOW_DEBUG
// these are undefined if SHOW_DEBUG is undefined,
// otherwise may be defined or not as needed
#define  BUTTON_PRESS_DEBUG
#define  SLIDER_DEBUG
#define  SLIDER_VAL_DEBUG
#undef   DEBUG_CMD_OVERLAP
#endif

// misc
#define  FOOTSWITCH_DEBOUNCE  50  // time in milliseconds

// variables and flags
boolean  main_screen;
boolean main_screen_active;
boolean alt_screen_active;
byte footswitch_state;
byte last_footswitch_state;
boolean trigger_set;
boolean trigger_holdoff;
byte pulse_duty_cycle;
int pulse_duration;
char buffer[BUFF_SIZE];
int  buffer_index;
char  replies_req;
boolean message_ready;
boolean serial_error;

// function prototypes
void draw_screen(void);
void draw_main_screen(void);
void draw_alt_screen(void);
void clear_screen(void);
void draw_button(int index, int button_x, int button_y, char *text);
void draw_slider_label(int label_x, int label_y, char *text);
void draw_slider(int index, int slider_x, int slider_y);
int decode_ASCII(char buffer[], int *index, char delim);
void print_pulse_profile(void);
void handle_button(void);
void handle_slider(void);
void do_pulse(void);
void handle_footswitch(void);
void wait_for_display(void);
void DISPLAY_SERIALEVENT(void);


/*
 * setup:
 *
 * the required Arduino sketch setup() function.
 */
 void setup()
{

  // timer2 output pin
  pinMode(PWM_OUT, OUTPUT);

  // footswitch input, with internal 20k pull-up
  pinMode(FOOTSWITCH_IN, INPUT);
  digitalWrite(FOOTSWITCH_IN, HIGH);

  // debug LED
  pinMode(DEBUG_LED, OUTPUT);

  // init the serial port and send a few <return>s
  // to get the display's attention
  DISPLAY_SIO.begin(115200);
  DISPLAY_SIO.print('\r');
  DISPLAY_SIO.print('\r');
  DISPLAY_SIO.print('\r');

#ifdef  SHOW_DEBUG
  DEBUG_SIO.begin(9600);
  DEBUG_SIO.println("Showing debug");
#endif

  // set our initial state to the main screen
  main_screen = true;
  main_screen_active = false;
  alt_screen_active = false;
  footswitch_state = digitalRead(FOOTSWITCH_IN);
  last_footswitch_state = footswitch_state;
  trigger_set = false;
  trigger_holdoff = false;
  pulse_duty_cycle = MIN_DUTY_CYCLE;
  pulse_duration = MIN_DURATION;
  buffer_index = 0;
  replies_req = 0;
  message_ready = false;
  serial_error = false;

  // set the font
  DISPLAY_SIO.println("f 13B");
  replies_req++;
  wait_for_display();

  // soften the button beep
  DISPLAY_SIO.println("bvs 25");
  replies_req++;
  wait_for_display();

}


/*
 * loop:
 *
 * the required Arduino loop() function.
 */
void loop()
{
  int  i, num_chars;
  
  draw_screen();

  // check for button and slider notifications
  if (message_ready)
  {
//    DEBUG_SIO.println("Handling a message");
    message_ready = false;
    if (buffer[0] == BUTTON_PRESS)
    {
      handle_button();
    }  /* button press */
    else if (buffer[0] == SLIDER_ACTION)
    {
      handle_slider();
    }  /* slider value */
    else
    {
      // unknown message - ignore
    }
  }

  // check for footswitch trigger
  handle_footswitch();

  // trigger pulse if flag is set
  if (trigger_set)
  {
    print_pulse_profile();
    trigger_set = false;
    do_pulse();
  }

}

/*
 * draw_screen:
 *
 * called to draw the proper screen based
 * on the main_screen state
 */
void draw_screen(void)
{

  if (main_screen)
  {
    // but only draw the screen if it is not yet active
    if (!main_screen_active)
    {
      draw_main_screen();
      // track the active screen
      main_screen_active = true;
      alt_screen_active = false;
    }
  }
  else
  {
    if (!alt_screen_active)
    {
      draw_alt_screen();
      alt_screen_active = true;
      main_screen_active = false;
    }
  }

}


/*
 * draw_main_screen:
 *
 * draws the buttons for the main screen.
 */
void draw_main_screen(void)
{

  clear_screen();

  // draw buttons with centered text
  // NOTE: button number assignments are also 'known'
  // in handle_button() switch statement
  draw_button(1, BUTTON_COL_1, BUTTON_ROW_1, "20%\n50ms");
  draw_button(2, BUTTON_COL_2, BUTTON_ROW_1, "20%\n100ms");
  draw_button(3, BUTTON_COL_3, BUTTON_ROW_1, "20%\n200ms");
  draw_button(4, BUTTON_COL_1, BUTTON_ROW_2, "40%\n50ms");
  draw_button(5, BUTTON_COL_2, BUTTON_ROW_2, "40%\n100ms");
  draw_button(6, BUTTON_COL_3, BUTTON_ROW_2, "40%\n200ms");
  draw_button(7, BUTTON_COL_1, BUTTON_ROW_3, "60%\n50ms");
  draw_button(8, BUTTON_COL_2, BUTTON_ROW_3, "60%\n100ms");
  draw_button(9, BUTTON_COL_3, BUTTON_ROW_3, "60%\n200ms");
  draw_button(10, BUTTON_COL_1, BUTTON_ROW_4, "99%\n50ms");
  draw_button(11, BUTTON_COL_2, BUTTON_ROW_4, "99%\n100ms");
  draw_button(12, BUTTON_COL_3, BUTTON_ROW_4, "99%\n200ms");

  draw_button(TRIGGER_BUTTON, BUTTON_COL_6, BUTTON_ROW_1, "TRIG");
  draw_button(ALT_SCREEN_BUTTON, BUTTON_COL_6, BUTTON_ROW_4, "AltScr");

}


/*
 * draw_alt_screen:
 *
 * draws the sliders and buttons for the alternate screen.
 */
void draw_alt_screen(void)
{

  clear_screen();

  draw_slider_label(DUTYCYCLE_SLIDER_LABEL_COL,
                    SLIDER_LABEL_ROW,
                    "Duty Cycle");
  draw_dutycycle_slider(DUTYCYCLE_SLIDER,
                    DUTYCYCLE_SLIDER_COL, SLIDER_ROW);

  draw_slider_label(DURATION_SLIDER_LABEL_COL,
                    SLIDER_LABEL_ROW,
                    "Duration");
  draw_duration_slider(DURATION_SLIDER,
                    DURATION_SLIDER_COL, SLIDER_ROW);

  draw_button(TRIGGER_BUTTON, BUTTON_COL_6, BUTTON_ROW_1, "TRIG");
  draw_button(MAIN_SCREEN_BUTTON, BUTTON_COL_6, BUTTON_ROW_4, "MainScr");

}


/*
 * clear_screen:
 *
 * called to clear the screen
 */
void clear_screen(void)
{

  wait_for_display();
  DISPLAY_SIO.println("z");
  replies_req++;

}


/*
 * draw_button:
 *
 * draws a 'big_button' at button_x,button_y with the label
 * 'text' centered in the button.  button is configured as
 * a Type 1 button (notify on press only).
 */
void draw_button(int index, int button_x, int button_y, char *text)
{

  wait_for_display();
  DISPLAY_SIO.print("bdc ");        // button with centered text
  DISPLAY_SIO.print(index, DEC);
  DISPLAY_SIO.print(" ");
  DISPLAY_SIO.print(button_x, DEC);
  DISPLAY_SIO.print(" ");
  DISPLAY_SIO.print(button_y, DEC);
  DISPLAY_SIO.print(" 1 \"");      // type 1 - notify on press only
  DISPLAY_SIO.print(text);
  DISPLAY_SIO.println("\" 1 2");   // 'big_button', formerly bitmaps 30 and 31
  replies_req++;

}


/*
 * draw_slider_label:
 *
 * called to draw a label for a slider
 */
void draw_slider_label(int label_x, int label_y, char *text)
{

  wait_for_display();
  DISPLAY_SIO.print("t \"");
  DISPLAY_SIO.print(text);
  DISPLAY_SIO.print("\" ");
  DISPLAY_SIO.print(label_x, DEC);
  DISPLAY_SIO.print(" ");
  DISPLAY_SIO.println(label_y, DEC);
  replies_req++;

}


/*
 * draw_dutycycle_slider:
 *
 * called to draw a DutyCycleSlider at slider_x
 * and slider_y.  knob is offset 5 pixels from edge.
 * lo/hi values are set to 1 and 99 to avoid 0 (and
 * needing to check for 0 in calc's).  The 'lo' vaue
 * is at the bottom of the slider.
 */
void draw_dutycycle_slider(int index, int slider_x, int slider_y)
{

  wait_for_display();
  DISPLAY_SIO.print("sl ");
  DISPLAY_SIO.print(index, DEC);
  DISPLAY_SIO.print(" 3 ");
  DISPLAY_SIO.print(slider_x, DEC);
  DISPLAY_SIO.print(" ");
  DISPLAY_SIO.print(slider_y, DEC);
  DISPLAY_SIO.println(" 5 7 0 1 1 99 0");
  replies_req++;

}


/*
 * draw_duration_slider:
 *
 * called to draw a DurationSlider at slider_x
 * and slider_y.  knob is offset 5 pixels from edge.
 * lo/hi values are set to 50 and 200.  The 'lo'
 * vaue is at the bottom of the slider.
 */
void draw_duration_slider(int index, int slider_x, int slider_y)
{

  wait_for_display();
  DISPLAY_SIO.print("sl ");
  DISPLAY_SIO.print(index, DEC);
  DISPLAY_SIO.print(" 4 ");
  DISPLAY_SIO.print(slider_x, DEC);
  DISPLAY_SIO.print(" ");
  DISPLAY_SIO.print(slider_y, DEC);
  DISPLAY_SIO.println(" 5 7 0 1 1 200 50");
  replies_req++;

}


/*
 * decode_ASCII:
 *
 * called to decod the ASCII numeric value at buffer[index].
 * returns the decoded value, and leaves index set to the
 * delimiter or the zero terminator.
 *
 * assumes that there are no more than three ASCII characters.
 */
int decode_ASCII(char buffer[], int *index, char delim)
{
  int  val, l_index;

  l_index = *index;
  val = buffer[l_index++] - '0';
  if ((buffer[l_index] != delim) && (buffer[l_index] != 0))
  {
    val = (val * 10) + (buffer[l_index++] - '0');
  }
  if ((buffer[l_index] != delim) && (buffer[l_index] != 0))
  {
    val = (val * 10) + (buffer[l_index++] - '0');
  }

#ifdef  PULSE_PROFILE_DEBUG
  DEBUG_SIO.print("decode_ASCII: ");
  DEBUG_SIO.print(val, DEC);
  DEBUG_SIO.print(" ending index: ");
  DEBUG_SIO.println(l_index, DEC);
#endif

  *index = l_index;
  return(val);

}


/*
 * print_pulse_profile:
 *
 * called to print the current low-level pulse profile
 */
void print_pulse_profile(void)
{

  wait_for_display();
  DISPLAY_SIO.print("t \"dc: ");
  DISPLAY_SIO.print(pulse_duty_cycle, DEC);
  DISPLAY_SIO.print("     \" ");  // extra spaces clear old text
  DISPLAY_SIO.print(BUTTON_COL_6, DEC);
  DISPLAY_SIO.print(" ");
  DISPLAY_SIO.println(BUTTON_ROW_3-20, DEC);
  replies_req++;

  wait_for_display();
  DISPLAY_SIO.print("t \"time: ");
  DISPLAY_SIO.print(pulse_duration, DEC);
  DISPLAY_SIO.print("     \" ");  // extra spaces clear old text
  DISPLAY_SIO.print(BUTTON_COL_6, DEC);
  DISPLAY_SIO.print(" ");
  DISPLAY_SIO.println(BUTTON_ROW_3, DEC);
  replies_req++;

}


/*
 * handle_button:
 *
 * handles button notifications for main screen and for
 * alt screen.  decodes button value into one of several
 * pre-set pulse profile settings, or into one of a few
 * special actions.
 */
void handle_button(void)
{
  int  index, button_num;
  
  // convert button number from ASCII
  index = 1;
  button_num = decode_ASCII(buffer, &index, ',');
#ifdef  BUTTON_PRESS_DEBUG
  DEBUG_SIO.print("button ");
  DEBUG_SIO.print(button_num, DEC);
  DEBUG_SIO.println(" pressed");
#endif

  switch (button_num)
  {
     case 1:
       pulse_duty_cycle = DUTY_CYCLE_20PCNT;
       pulse_duration = 50;   // ms
       break;
     case 2:
       pulse_duty_cycle = DUTY_CYCLE_20PCNT;
       pulse_duration = 100;  // ms
       break;
     case 3:
       pulse_duty_cycle = DUTY_CYCLE_20PCNT;
       pulse_duration = 200;  // ms
       break;
     case 4:
       pulse_duty_cycle = DUTY_CYCLE_40PCNT;
       pulse_duration = 50;   // ms
       break;
     case 5:
       pulse_duty_cycle = DUTY_CYCLE_40PCNT;
       pulse_duration = 100;  // ms
       break;
     case 6:
       pulse_duty_cycle = DUTY_CYCLE_40PCNT;
       pulse_duration = 200;  // ms
       break;
     case 7:
       pulse_duty_cycle = DUTY_CYCLE_60PCNT;
       pulse_duration = 50;   // ms
       break;
     case 8:
       pulse_duty_cycle = DUTY_CYCLE_60PCNT;
       pulse_duration = 100;  // ms
       break;
     case 9:
       pulse_duty_cycle = DUTY_CYCLE_60PCNT;
       pulse_duration = 200;  // ms
       break;
     case 10:
       pulse_duty_cycle = DUTY_CYCLE_99PCNT;
       pulse_duration = 50;   // ms
       break;
     case 11:
       pulse_duty_cycle = DUTY_CYCLE_99PCNT;
       pulse_duration = 100;  // ms
       break;
     case 12:
       pulse_duty_cycle = DUTY_CYCLE_99PCNT;
       pulse_duration = 200;  // ms
       break;

     case MAIN_SCREEN_BUTTON:
       main_screen = true;
       break;
     case ALT_SCREEN_BUTTON:
       main_screen = false;
       break;
     case TRIGGER_BUTTON:    // same on both screens
       trigger_set = true;
       break;

     default:
#ifdef  BUTTON_PRESS_DEBUG
       DEBUG_SIO.print("OOPS - unknown button ");
       DEBUG_SIO.print(button_num, DEC);
       DEBUG_SIO.println(" pressed");
#endif
       pulse_duty_cycle = MIN_DUTY_CYCLE;
       pulse_duration = MIN_DURATION;
       break;
  }

}


/*
 * handle_slider:
 *
 * handles slider notifications for the alt screen.  decodes
 * slider number and value, then maps slider value into the
 * appropriate range for the slider function.
 */
void handle_slider(void)
{
  int  index, slider_num, slider_val;
    
  // convert slider number from ASCII
  index = 1;
  slider_num = decode_ASCII(buffer, &index, ':');
  index++;    // skip delimiter
  slider_val = decode_ASCII(buffer, &index, ':');
#ifdef  SLIDER_VAL_DEBUG
  // NOTE: slow response with this enabled.
  DEBUG_SIO.print("Slider: ");
  DEBUG_SIO.print(slider_num, DEC);
  DEBUG_SIO.print(" val: ");
  DEBUG_SIO.println(slider_val, DEC);
#endif
  if (slider_num == DUTYCYCLE_SLIDER)
  {
    // the dutycycle slider sends a percentage 1-99
    pulse_duty_cycle = map(slider_val,
                            0, 100,
                            MIN_DUTY_CYCLE, DUTY_CYCLE_MAX-1);
  }
  else if (slider_num == DURATION_SLIDER)
  {
    // the duraction slider sends a time in milliseconds
    if (slider_val < MIN_DURATION)
    {
      pulse_duration = MIN_DURATION;
    }
    else if (slider_val > MAX_DURATION)
    {
      pulse_duration = MAX_DURATION;
    }
    else pulse_duration = slider_val;  // OK to use directly
  }
  else
  {
    pulse_duty_cycle = MIN_DUTY_CYCLE;
    pulse_duration = MIN_DURATION;
#ifdef  SLIDER_DEBUG
    DEBUG_SIO.print("Unknown slider: ");
    DEBUG_SIO.println(slider_num, DEC);
#endif
  }

}


/*
 * do_pulse:
 *
 * called to issue a pulse as specified by the current
 * pulse profile settings.
 *
 * NOTE: The direct use of processor registers makes this code
 * specific to a processor function (output OSC2B) instead of
 * an Arduino function (e.g. analogWrite on pin 3) and this
 * processor function may appear on different Arduino pins
 * depending on which processor is used.  The ifdefs at the
 * beginning of this file allow this same code to be used on a
 * Uno and on a Mega 2560.
 */
void do_pulse(void)
{

  digitalWrite(DEBUG_LED, 1);

  // set timer2 for fast PWM, start timer
  TCCR2A = _BV(COM2A1) | _BV(COM2B1) | _BV(WGM21) | _BV(WGM20);
  TCCR2B = _BV(CS20);  // prescler = 1

  // the pulse
  OCR2B = pulse_duty_cycle;
  delay(pulse_duration);

  OCR2B = 1;          // minimum duty cycle
  TCCR2A = _BV(0);    // stop timer

  digitalWrite(DEBUG_LED, 0);

}


/*
 * handle_footswitch:
 *
 * called to monitor and debounce the footswitch input.
 * sets trigger_set when a valid footswitch action is
 * detected (falling edge).  waits for rising edge to
 * reenable footswitch action.
 */
void handle_footswitch(void)
{

  footswitch_state = digitalRead(FOOTSWITCH_IN);
  if (footswitch_state != last_footswitch_state)
  {
    if ((footswitch_state == LOW)  && (!trigger_holdoff))
    {
      delay(FOOTSWITCH_DEBOUNCE);
      footswitch_state = digitalRead(FOOTSWITCH_IN);
      // if footswitch is still low, OK to trigger pulse
      if (footswitch_state == LOW)
      {
        trigger_set = true;
        trigger_holdoff = true;
      }
    }
    last_footswitch_state = footswitch_state;
  }
  else
  {
    // fotswitch has been released, allow new trigger
    if (footswitch_state == HIGH)
    {
      trigger_holdoff = false;
    }
  }

}


/*
 * wait_for_display:
 *
 * called to wait for display to reply to previous commands.
 * used in lieu of XON/XOFF flow control.  application is able
 * to send up to MAX_CMD_OUTSTANDING overlapping commands before
 * needing to wait for display to reply.
 */
void wait_for_display(void)
{

  while (replies_req >= MAX_CMD_OUTSTANDING)
  {
#ifdef  DEBUG_CMD_OVERLAP
    DEBUG_SIO.print("Stalled: ");
    DEBUG_SIO.println(replies_req, DEC);
#endif
    // wait until display catches up with commands
    DISPLAY_SERIALEVENT();  // must call manually
    delay(10);
  }

}


/*
 * DISPLAY_SERIALEVENT:
 *
 * called when serial data is available.  reads serial data into
 * the buffer and sets the line_ready flag when a <return> char
 * has been received.
 *
 * NOTE: This function is called by the Arduino environment each
 * time loop() is executed, after loop() is called.  If loop()
 * is waiting for serial data it will need to call this function
 * manually.
 */
void DISPLAY_SERIALEVENT(void)
{
  char  serial_char;

  while (DISPLAY_SIO.available())
  {
    serial_char = DISPLAY_SIO.read();
  
#undef  DEBUG_SERIAL
#ifdef  DEBUG_SERIAL
    // send same data to debug serial port
    DEBUG_SIO.print(serial_char);
    if (serial_char == '\r')
    {
      DEBUG_SIO.print("<return>");
      DEBUG_SIO.print("\n");
    }
#endif

    if (serial_char == '>')
    {
      if (replies_req)
      {
        replies_req--;
      }
    }
    // lines starting with:
    //   '!' indicates command failure
    //   '^' indicates display buffer full
    //   '?' indicates transmission line error
    //   '#' indicates CRC error
    //   ':' indicates human-readable information
    else if ((buffer_index == 0) &&
              ((serial_char == '!') ||
               (serial_char == '^') ||
               (serial_char == '?') ||
               (serial_char == '#') ||
               (serial_char == ':')))
    {
      serial_error = true;
      // additional handling depends on user application
    }
    else if (serial_char == '\r')
    {
      // end of line
      buffer[buffer_index] = '\0';  // string terminator
      if (buffer_index > 0)
      {
#ifdef  DEBUG_SERIAL
        DEBUG_SIO.println("Got a message");
#endif
        message_ready = true;
        buffer_index = 0;
      }
    }
    else
    {
      // store character
      buffer[buffer_index] = serial_char;
      buffer_index++;
    }
    
  }  /* while */

}


